from inspect import isclass
from typing import List, Tuple

from .base import GenericModelCodeGenerator, KWAGRS_TEMPLATE, METADATA_FIELD_NAME, sort_kwargs, template
from ..dynamic_typing import DDict, DList, DOptional, ImportPathList, MetaData, ModelMeta, StringLiteral, StringSerializable

DEFAULT_ORDER = (
    ("default", "converter", "factory"),
    "*",
    ("metadata",)
)


class AttrsModelCodeGenerator(GenericModelCodeGenerator):
    ATTRS = template(f"attr.s{{% if kwargs %}}({KWAGRS_TEMPLATE}){{% endif %}}")
    ATTRIB = template(f"attr.ib({KWAGRS_TEMPLATE})")
    default_types_style = {
        StringLiteral: {
            StringLiteral.TypeStyle.use_literals: False
        }
    }

    def __init__(self, model: ModelMeta, meta=False, attrs_kwargs: dict = None, **kwargs):
        """
        :param model: ModelMeta instance
        :param meta: Enable generation of metadata as attrib argument
        :param attrs_kwargs: kwargs for @attr.s() decorators
        :param kwargs:
        """
        super().__init__(model, **kwargs)
        self.no_meta = not meta
        self.attrs_kwargs = attrs_kwargs or {}

    @property
    def decorators(self) -> Tuple[ImportPathList, List[str]]:
        imports, decorators = super().decorators
        imports.append(('attr', None))
        decorators.insert(0, self.ATTRS.render(kwargs=self.attrs_kwargs))
        return imports, decorators

    def field_data(self, name: str, meta: MetaData, optional: bool) -> Tuple[ImportPathList, dict]:
        """
        Form field data for template

        :param name: Original field name
        :param meta: Field metadata
        :param optional: Is field optional
        :return: imports, field data
        """
        imports, data = super().field_data(name, meta, optional)
        body_kwargs = {}
        if optional:
            meta: DOptional
            if isinstance(meta.type, DList):
                body_kwargs["factory"] = "list"
            elif isinstance(meta.type, DDict):
                body_kwargs["factory"] = "dict"
            else:
                body_kwargs["default"] = "None"
                if isclass(meta.type) and issubclass(meta.type, StringSerializable) and not self.post_init_converters:
                    body_kwargs["converter"] = f"optional({meta.type.__name__})"
                    imports.append(("attr.converter", "optional"))
        elif isclass(meta) and issubclass(meta, StringSerializable) and not self.post_init_converters:
            body_kwargs["converter"] = meta.__name__

        if not self.no_meta and name != data["name"]:
            body_kwargs["metadata"] = {METADATA_FIELD_NAME: name}
        data["body"] = self.ATTRIB.render(kwargs=sort_kwargs(body_kwargs, DEFAULT_ORDER))
        return imports, data

    @property
    def convert_strings_kwargs(self) -> Tuple[ImportPathList, dict]:
        """
        :return: Imports and Dict with kw-arguments for `json_to_models.models.string_converters.convert_strings` decorator.
        """
        imports, kwargs = super().convert_strings_kwargs
        imports.append(('json_to_models.models', ['ClassType']))
        kwargs["class_type"] = 'ClassType.Attrs'
        return imports, kwargs



from itertools import permutations
from typing import ClassVar, Collection, Dict, Iterable, List, Set, Tuple, Type, Union

from .base import BaseType, ImportPathList


class StringSerializable(BaseType):
    """
    Mixin for classes which are used to (de-)serialize some values in a string form
    """

    class TypeStyle:
        use_actual_type = 'use_actual_type'

    actual_type: ClassVar[Type]

    @classmethod
    def to_internal_value(cls, value: str) -> 'StringSerializable':
        """
        Factory method

        :raises ValueError: if this class can not represent given value
        :param value: some string literal
        :return: Instance of this class
        """
        raise NotImplementedError()

    def to_representation(self) -> str:
        """
        Convert instance to string literal

        :return: string literal
        """
        raise NotImplementedError()

    @classmethod
    def to_typing_code(cls, types_style: Dict[Union['BaseType', Type['BaseType']], dict]) -> Tuple[ImportPathList, str]:
        """
        Unlike other BaseType's subclasses it's a class method because StringSerializable instance is not parameterized
        as a metadata instance but contains actual data
        """
        cls_name = cls.__name__
        options = cls.get_options_for_type(cls, types_style)
        if options.get(cls.TypeStyle.use_actual_type):
            if cls.actual_type.__module__ != 'builtins':
                return [(cls.actual_type.__module__, cls.actual_type.__name__)], cls.actual_type.__name__
            return [], cls.actual_type.__name__
        return [('json_to_models.dynamic_typing', cls_name)], cls_name

    def __iter__(self):
        return iter(())


T_StringSerializable = Type[StringSerializable]


class StringSerializableRegistry:
    def __init__(self, *types: T_StringSerializable):
        self.types: List[T_StringSerializable] = list(types)
        self.replaces: Set[Tuple[T_StringSerializable, T_StringSerializable]] = set()

    def __iter__(self):
        return iter(self.types)

    def __contains__(self, item):
        return item in self.types

    def add(self, replace_types: Iterable[T_StringSerializable] = (), cls: type = None):
        """
        Register decorated class in registry. Can be called as a method if cls argument is passed.

        :param replace_types: List of classes that is the particular case of decorated one
        :param cls: StringSerializable class
        :return: decorator
        """

        def decorator(cls):
            self.types.append(cls)
            for t in replace_types:
                self.replaces.add((t, cls))
            return cls

        if cls:
            decorator(cls)
            return

        return decorator

    def remove(self, cls: type):
        """
        Unregister given class

        :param cls: StringSerializable class
        """
        self.types.remove(cls)
        for base, replace in list(self.replaces):
            if replace is cls or base is cls:
                self.replaces.remove((base, replace))

    def resolve(self, *types: T_StringSerializable) -> Collection[T_StringSerializable]:
        """
        Return set of StringSerializable classes which can represent all classes from types argument.

        :param types: Sequence of StringSerializable classes
        :return: Set of StringSerializable
        """
        # TODO: Resolve common type of 2 different types (e.g str from float and bool)
        # Do it by getting all childs of each class with their level then merge it into one list and find one with min(max(levels) for c n childs)
        types = set(types)
        flag = True
        while flag:
            flag = False
            filtered: Set[T_StringSerializable] = set()
            for t1, t2 in permutations(types, 2):
                if (t1, t2) in self.replaces:
                    filtered.add(t2)
                    flag = True
            if flag:
                types = filtered
        # noinspection PyUnboundLocalVariable
        return types


# Default registry
registry = StringSerializableRegistry()


@registry.add()
class IntString(StringSerializable, int):
    actual_type = int

    @classmethod
    def to_internal_value(cls, value: str) -> 'IntString':
        return cls(value)

    def to_representation(self) -> str:
        return str(self)


@registry.add(replace_types=(IntString,))
class FloatString(StringSerializable, float):
    actual_type = float

    @classmethod
    def to_internal_value(cls, value: str) -> 'FloatString':
        return cls(value)

    def to_representation(self) -> str:
        return str(self)


@registry.add()
class BooleanString(StringSerializable, int):
    # We can't extend bool class, but we can extend int with same result excepting isinstance and issubclass check
    actual_type = bool

    @classmethod
    def to_internal_value(cls, value: str) -> 'BooleanString':
        b = {"true": True, "false": False}.get(value.lower(), None)
        if b is None:
            raise ValueError(f"invalid literal for bool: '{value}'")
        return cls(b)

    def to_representation(self) -> str:
        return str(bool(self)).lower()



from typing import Dict, List, Type, Union

import pytest
from typing_extensions import Literal

from json_to_models.dynamic_typing import (AbsoluteModelRef, BaseType, DDict, DList, DOptional, IntString, IsoDateString, ModelMeta, ModelPtr, StringLiteral, StringSerializable, Unknown, compile_imports)
from json_to_models.models.base import GenericModelCodeGenerator, generate_code
from json_to_models.models.structure import sort_fields
from json_to_models.models.utils import indent

LITERAL_SOURCE = f"from {Literal.__module__}"

# Data structure:
# (string, indent lvl, indent string)
# result
test_indent_data = [
    pytest.param(
        ("1", 1, " " * 4),
        "    1"
    ),
    pytest.param(
        ("1\n2", 1, " " * 4),
        "    1\n    2"
    ),
    pytest.param(
        ("1\n2", 2, " " * 4),
        "        1\n        2"
    ),
    pytest.param(
        ("1\n    2", 2, " " * 4),
        "        1\n            2"
    ),
]


@pytest.mark.parametrize("args,expected", test_indent_data)
def test_indent(args, expected):
    assert indent(*args) == expected


def model_factory(name: str, metadata: dict):
    model = ModelMeta(metadata, name)
    model.set_raw_name(name)
    return model


INDENT = " " * 4 * 2


def trim(s: str):
    if "\n" in s:
        n = len(INDENT)
        lines = s.split("\n")
        for i in (0, -1):
            if not lines[i].strip():
                del lines[i]

        s = "\n".join(line[n:] if line[:n] == INDENT else line for line in lines)
    return s


# Data structure:
# pytest.param id -> {
#   "model" -> (model_name, model_metadata),
#   test_name -> expected, ...
# }
test_data = {
    "base": {
        "model": ("Test", {
            "foo": int,
            "bar": int,
            "baz": float
        }),
        "fields_data": {
            "foo": {
                "name": "foo",
                "type": "int"
            },
            "bar": {
                "name": "bar",
                "type": "int"
            },
            "baz": {
                "name": "baz",
                "type": "float"
            }
        },
        "fields": {
            "imports": "",
            "fields": [
                "foo: int",
                "bar: int",
                "baz: float",
            ]
        },
        "generated": trim("""
        class Test:
            foo: int
            bar: int
            baz: float
        """)
    },
    "complex": {
        "model": ("Test", {
            "foo": int,
            "baz": DOptional(DList(DList(str))),
            "bar": IntString,
            "d": DDict(Unknown)
        }),
        "fields_data": {
            "foo": {
                "name": "foo",
                "type": "int"
            },
            "baz": {
                "name": "baz",
                "type": "Optional[List[List[str]]]"
            },
            "bar": {
                "name": "bar",
                "type": "IntString"
            },
            "d": {
                "name": "d",
                "type": "Dict[str, Any]"
            }
        },
        "fields": {
            "imports": "from json_to_models.dynamic_typing import IntString\n"
                       "from typing import Any, Dict, List, Optional",
            "fields": [
                "foo: int",
                "bar: IntString",
                "d: Dict[str, Any]",
                "baz: Optional[List[List[str]]]",
            ]
        },
        "generated": trim("""
        from json_to_models.dynamic_typing import IntString
        from typing import Any, Dict, List, Optional
        
        
        class Test:
            foo: int
            bar: IntString
            d: Dict[str, Any]
            baz: Optional[List[List[str]]]
        """)
    }
}

test_data_unzip = {
    test: [
        pytest.param(
            model_factory(*data["model"]),
            data[test],
            id=id
        )
        for id, data in test_data.items()
        if test in data
    ]
    for test in ("fields_data", "fields", "generated")
}


@pytest.mark.parametrize("value,expected", test_data_unzip["fields_data"])
def test_fields_data(value: ModelMeta, expected: Dict[str, dict]):
    gen = GenericModelCodeGenerator(value)
    required, optional = sort_fields(value)
    for is_optional, fields in enumerate((required, optional)):
        for field in fields:
            field_imports, data = gen.field_data(field, value.type[field], bool(is_optional))
            assert data == expected[field]


@pytest.mark.parametrize("value,expected", test_data_unzip["fields"])
def test_fields(value: ModelMeta, expected: dict):
    expected_imports: str = expected["imports"]
    expected_fields: List[str] = expected["fields"]
    gen = GenericModelCodeGenerator(value)
    imports, fields = gen.fields
    imports = compile_imports(imports)
    assert imports == expected_imports
    assert fields == expected_fields


@pytest.mark.parametrize("value,expected", test_data_unzip["generated"])
def test_generated(value: ModelMeta, expected: str):
    generated = generate_code(([{"model": value, "nested": []}], {}), GenericModelCodeGenerator)
    assert generated.rstrip() == expected, generated


def test_absolute_model_ref():
    test_model = ModelMeta({"field": int}, "A")
    test_model.name = "test_model"
    test_ptr = ModelPtr(test_model)
    assert test_ptr.to_typing_code({})[1] == "'TestModel'"
    with AbsoluteModelRef.inject({test_model: "Parent"}):
        assert test_ptr.to_typing_code({})[1] == "'Parent.TestModel'"
    assert test_ptr.to_typing_code({})[1] == "'TestModel'"
    with AbsoluteModelRef.inject({test_model: "Parent"}):
        assert test_ptr.to_typing_code({})[1] == "'Parent.TestModel'"
        with AbsoluteModelRef.inject({test_model: "AnotherParent"}):
            assert test_ptr.to_typing_code({})[1] == "'AnotherParent.TestModel'"
        assert test_ptr.to_typing_code({})[1] == "'Parent.TestModel'"

    wrapper = DList(DList(test_ptr))
    assert wrapper.to_typing_code({})[1] == "List[List['TestModel']]"
    with AbsoluteModelRef.inject({test_model: test_model}):
        assert wrapper.to_typing_code({})[1] == "List[List['TestModel.TestModel']]"


test_unicode_data = [
    pytest.param(
        model_factory("Test", {
            "foo": int,
            "bar": int,
            "baz": float
        }),
        {'convert_unicode': True},
        trim("""
        class Test:
            foo: int
            bar: int
            baz: float
        """),
        id="test_pytest_setup"
    ),
    pytest.param(
        model_factory("Test", {
            "поле1": int,
            "bar": int,
            "baz": float
        }),
        {'convert_unicode': True},
        trim("""
        class Test:
            pole1: int
            bar: int
            baz: float
        """),
        id="test_field_on"
    ),
    pytest.param(
        model_factory("Test", {
            "поле1": int,
            "bar": int,
            "baz": float
        }),
        {'convert_unicode': False},
        trim("""
        class Test:
            поле1: int
            bar: int
            baz: float
        """),
        id="test_field_off"
    ),
    pytest.param(
        model_factory("Тест", {
            "поле1": int,
            "bar": int,
            "baz": float
        }),
        {'convert_unicode': True},
        trim("""
        class Test:
            pole1: int
            bar: int
            baz: float
        """),
        id="test_field_on"
    ),
    pytest.param(
        model_factory("Тест", {
            "поле1": int,
            "bar": int,
            "baz": float
        }),
        {'convert_unicode': False},
        trim("""
        class Тест:
            поле1: int
            bar: int
            baz: float
        """),
        id="test_classname_off"
    ),
]


@pytest.mark.parametrize("value,kwargs,expected", test_unicode_data)
def test_unicode(value: ModelMeta, kwargs: dict, expected: str):
    generated = generate_code(([{"model": value, "nested": []}], {}),
                              GenericModelCodeGenerator, class_generator_kwargs=kwargs)
    assert generated.rstrip() == expected, generated


# Data format:
# (
#   model metadata,
#   style override,
#   expected
# )
test_override_style_data = [
    pytest.param(
        model_factory("M", {
            "bar": StringLiteral({'bar', 'foo'})
        }),
        {},
        trim(f"""
        {LITERAL_SOURCE} import Literal
        
        
        class M:
            bar: Literal["bar", "foo"]
        """),
        id='default_behaviour'
    ),
    pytest.param(
        model_factory("M", {
            "bar": StringLiteral({'bar', 'foo'})
        }),
        {StringLiteral: {
            StringLiteral.TypeStyle.use_literals: False
        }},
        trim("""
        class M:
            bar: str
        """),
        id='disable_literal'
    ),
    pytest.param(
        model_factory("M", {
            "bar": IntString
        }),
        {IntString: {
            IntString.TypeStyle.use_actual_type: True
        }},
        trim("""
        class M:
            bar: int
        """),
        id='string_serializable_use_actual_type'
    ),
    pytest.param(
        model_factory("M", {
            "bar": IntString
        }),
        {StringSerializable: {
            StringSerializable.TypeStyle.use_actual_type: True
        }},
        trim("""
        class M:
            bar: int
        """),
        id='string_serializable_use_actual_type_wildcard'
    ),
    pytest.param(
        model_factory("M", {
            "bar": IsoDateString
        }),
        {IsoDateString: {
            IsoDateString.TypeStyle.use_actual_type: True
        }},
        trim("""
        from datetime import date
        
        
        class M:
            bar: date
        """),
        id='string_serializable_use_actual_type_date'
    ),
]


@pytest.mark.parametrize("value,types_style,expected", test_override_style_data)
def test_override_style(value: ModelMeta, types_style: Dict[Union['BaseType', Type['BaseType']], dict], expected: str):
    generated = generate_code(
        ([{"model": value, "nested": []}], {}),
        GenericModelCodeGenerator,
        class_generator_kwargs=dict(types_style=types_style)
    )
    assert generated.rstrip() == expected, generated



from collections import defaultdict
from itertools import chain, combinations
from typing import Dict, FrozenSet, Iterable, List, Set, Tuple

from ordered_set import OrderedSet

from .dynamic_typing import BaseType, MetaData, ModelMeta, ModelPtr
from .utils import Index, distinct_words


class ModelCmp:
    """
    Generic model comparator
    """

    def cmp(self, fields_a: set, fields_b: set) -> bool:
        raise NotImplementedError()


class ModelFieldsEquals(ModelCmp):
    def cmp(self, fields_a: set, fields_b: set) -> bool:
        return fields_a == fields_b


class ModelFieldsPercentMatch(ModelCmp):
    DEFAULT = .7

    def __init__(self, percent_fields: float = DEFAULT):
        self.percent_fields = percent_fields

    def cmp(self, fields_a: set, fields_b: set) -> bool:
        return len(fields_a & fields_b) / len(fields_a | fields_b) >= self.percent_fields


class ModelFieldsNumberMatch(ModelCmp):
    DEFAULT = 10

    def __init__(self, number_fields: int = DEFAULT):
        self.number_fields = number_fields

    def cmp(self, fields_a: set, fields_b: set) -> bool:
        return len(fields_a & fields_b) >= self.number_fields


class ModelRegistry:
    DEFAULT_MODELS_CMP = (ModelFieldsPercentMatch(), ModelFieldsNumberMatch())

    def __init__(self, *models_cmp: ModelCmp):
        """
        :param models_cmp: list of model comparators. If you want merge only equals models pass ModelFieldsEquals()
        """
        self._models_cmp = models_cmp or self.DEFAULT_MODELS_CMP
        self._registry: Dict[str, ModelMeta] = {}
        self._index = Index()

    @property
    def models(self):
        return self._registry.values()

    @property
    def models_map(self):
        return self._registry

    def process_meta_data(
            self, meta: MetaData,
            model_name: str = None,
            parent: MetaData = None,
            parent_model: Tuple[ModelMeta, str] = (None, None),
            replace_kwargs=None
    ) -> ModelPtr:
        """
        Convert metadata (dict) to model and return a pointer to this model

        :param meta: Dict[str, MetaData] - usually result of MetadataGenerator.generate method call
        :param model_name: Raw model name. Will be set as is.
            It is recommended to set it for root level models because names can not be generated for such models
        :param parent: System attribute for recursive call
        :param parent_model: Pair of (ModelMeta, field name). System attribute for recursive call
        :param replace_kwargs: System attribute for recursive call
        :return:
        """
        replace_kwargs = replace_kwargs or {}
        ptr = None

        if isinstance(meta, dict):
            # Register model
            model_meta = self._register(meta)
            ptr = ModelPtr(model_meta, parent=parent_model[0], parent_field_name=parent_model[1])
            if parent:
                parent.replace(ptr, **replace_kwargs)

            # Process nested data
            for key, value in meta.items():
                nested_ptr = self.process_meta_data(value, parent_model=(model_meta, key))
                if nested_ptr:
                    meta[key] = nested_ptr

        elif isinstance(meta, BaseType):
            # Process other non-atomic types
            try:
                meta_iter = iter(meta)
            except TypeError:
                pass
            else:
                for i, nested_meta in enumerate(meta_iter):
                    self.process_meta_data(
                        nested_meta,
                        parent=meta,
                        parent_model=parent_model,
                        replace_kwargs={'parent': meta, 'index': i}
                    )

        if model_name is not None:
            ptr.type.set_raw_name(model_name)
        return ptr

    def _register(self, meta: MetaData):
        model_meta = ModelMeta(meta, self._index()) if not isinstance(meta, ModelMeta) else meta
        self._registry[model_meta.index] = model_meta
        return model_meta

    def _unregister(self, model_meta: ModelMeta):
        del self._registry[model_meta.index]

    def _models_cmp_fn(self, model_a: ModelMeta, model_b: ModelMeta) -> bool:
        """
        Return True if ANY of comparators decided to merge pair of models
        """
        fields_a = set(model_a.type.keys())
        fields_b = set(model_b.type.keys())
        return any(cmp.cmp(fields_a, fields_b) for cmp in self._models_cmp)

    def merge_models(self, generator, strict=False) -> List[Tuple[ModelMeta, Set[ModelMeta]]]:
        """
        Optimize whole models registry by merging same or similar models (uses given models comparators)

        :param generator: Generator instance that will be used to metadata merging and optimization
        :param strict: if True ALL models in merge group should meet the conditions
            else groups will form from pairs of models as is.
        :return: pairs of (new model, set of old models)
        """
        # TODO: Implement strict mode
        models2merge: Dict[ModelMeta, Set[ModelMeta]] = defaultdict(set)
        for model_a, model_b in combinations(self.models, 2):
            if self._models_cmp_fn(model_a, model_b):
                models2merge[model_a].add(model_b)
                models2merge[model_b].add(model_a)

        # Groups of models to merge
        groups: Iterable[Set[ModelMeta]] = [{model, *models} for model, models in models2merge.items()]
        # Make groups non-overlapping.
        # This is not optimal algorithm but it works and we probably will not have thousands of models here.
        flag = True
        while flag:
            flag = False
            new_groups: OrderedSet[FrozenSet[ModelMeta]] = OrderedSet()
            for gr1 in groups:
                in_set = False
                for gr2 in groups:
                    if gr1 is gr2:
                        continue
                    if gr1 & gr2:
                        in_set = True
                        old_len = len(new_groups)
                        new_groups.add(frozenset(gr1 | gr2))
                        added = old_len < len(new_groups)
                        flag = flag or added
                if not in_set:
                    new_groups.add(gr1)
            if flag:
                groups: OrderedSet[FrozenSet[ModelMeta]] = new_groups

        replaces = []
        replaces_ids = set()
        for group in groups:
            model_meta = self._merge(generator, *group)
            generator.optimize_type(model_meta)
            replaces_ids.add(model_meta.index)
            replaces.append((model_meta, group))

        for model_meta in self.models:
            generator.optimize_type(model_meta)
        return replaces

    def _merge(self, generator, *models: ModelMeta):
        """
        Merge models by generator.merge_field_sets method.
        Also ensure that the new model will have all metadata from old ones
        """
        original_fields = list(chain(model.original_fields for model in models))
        originals_names = []
        for model in models:
            if not model.is_name_generated and model.name:
                originals_names.append(model.name)
        originals_names = distinct_words(*originals_names)

        metadata = generator.merge_field_sets([model.type for model in models])
        model_meta = ModelMeta(metadata, self._index(), original_fields)
        if originals_names:
            model_meta.set_raw_name(ModelMeta.name_joiner(*sorted(originals_names)))
        for model in models:
            self._unregister(model)
            for ptr in tuple(model.pointers):
                ptr.replace(model_meta)
            for ptr in tuple(model.child_pointers):
                ptr.replace_parent(model_meta)
        self._register(model_meta)

        return model_meta

    def fix_name_duplicates(self):
        """
        Add model's index to all duplicate names
        """
        counter = defaultdict(int)
        for model in self.models:
            counter[model.name or model.index] += 1
            if counter[model.name] > 1:
                model.set_raw_name(model.name_joiner(model.name, model.index), generated=True)

    def generate_names(self):
        for model in self.models:
            if model.is_name_generated is None:
                model.generate_name()
            if model.name is None:
                model.set_raw_name("Unknown_" + model.index, True)
        self.fix_name_duplicates()



from typing import List, Optional, Tuple

from .base import GenericModelCodeGenerator, KWAGRS_TEMPLATE, sort_kwargs, template
from ..dynamic_typing import (
    DDict,
    DList,
    DOptional,
    ImportPathList,
    MetaData,
    ModelMeta,
    Null,
    StringLiteral,
    StringSerializable,
    Unknown
)

DEFAULT_ORDER = (
    "*",
)


class PydanticModelCodeGenerator(GenericModelCodeGenerator):
    PYDANTIC_FIELD = template("Field({{ default }}{% if kwargs %}, KWAGRS_TEMPLATE{% endif %})"
                              .replace('KWAGRS_TEMPLATE', KWAGRS_TEMPLATE))
    default_types_style = {
        StringSerializable: {
            StringSerializable.TypeStyle.use_actual_type: True
        },
        StringLiteral: {
            StringLiteral.TypeStyle.use_literals: True
        }
    }

    def __init__(self, model: ModelMeta, **kwargs):
        """
        :param model: ModelMeta instance
        :param kwargs:
        """
        kwargs['post_init_converters'] = False
        super().__init__(model, **kwargs)

    def generate(self, nested_classes: List[str] = None, extra: str = "", **kwargs) \
            -> Tuple[ImportPathList, str]:
        imports, body = super(PydanticModelCodeGenerator, self).generate(
            bases='BaseModel',
            nested_classes=nested_classes,
            extra=extra
        )
        imports.append(('pydantic', ['BaseModel', 'Field']))
        return imports, body

    def _filter_fields(self, fields):
        fields = super()._filter_fields(fields)
        filtered = []
        for field in fields:
            field_type = self.model.type[field]
            if field_type in (Unknown, Null):
                continue
            filtered.append(field)
        return filtered

    def field_data(self, name: str, meta: MetaData, optional: bool) -> Tuple[ImportPathList, dict]:
        """
        Form field data for template

        :param name: Original field name
        :param meta: Field metadata
        :param optional: Is field optional
        :return: imports, field data
        """
        imports, data = super().field_data(name, meta, optional)
        default: Optional[str] = None
        body_kwargs = {}
        if optional:
            meta: DOptional
            if isinstance(meta.type, DList):
                default = "[]"
            elif isinstance(meta.type, DDict):
                default = "{}"
            else:
                default = "None"

        if name != data["name"]:
            body_kwargs["alias"] = f'"{name}"'
        if body_kwargs:
            data["body"] = self.PYDANTIC_FIELD.render(
                default=default or '...',
                kwargs=sort_kwargs(body_kwargs, DEFAULT_ORDER)
            )
        elif default is not None:
            data["body"] = default
        return imports, data



import multiprocessing
import sys

from setuptools import find_packages, setup
from setuptools.command.test import test as TestCommand

import json_to_models

REPO = "bogdandm/json2python-models"
CPU_N = multiprocessing.cpu_count()

with open('requirements.txt') as f:
    required = f.read().splitlines()
with open('README.md') as f:
    long_description = f.read()
long_description = long_description.replace("/etc", f"https://raw.githubusercontent.com/{REPO}/master/etc")


class PyTest(TestCommand):
    user_options = [("pytest-args=", "a", "Arguments to pass to pytest")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = ""

    def run_tests(self):
        import shlex
        import pytest
        args = self.pytest_args
        if CPU_N > 1 and "-n " not in args:
            args += f" -n {CPU_N}"
        errno = pytest.main(shlex.split(args))
        sys.exit(errno)


setup(
    name="json2python-models",
    version=json_to_models.__version__,
    python_requires=">=3.7",
    url=f"https://github.com/{REPO}",
    author="bogdandm (Bogdan Kalashnikov)",
    author_email="bogdan.dm1995@yandex.ru",
    description="Python models (pydantic, attrs, dataclasses or custom) generator from JSON data with typing module support",
    long_description=long_description,
    long_description_content_type='text/markdown',
    license="MIT",
    packages=find_packages(exclude=['test', 'testing_tools']),
    entry_points={
        'console_scripts': ['json2models = json_to_models.cli:main']
    },
    install_requires=required,
    cmdclass={"test": PyTest},
    tests_require=["pytest>=4.4.0", "pytest-xdist", "requests", "attrs", "pydantic>=1.3", "ruamel.yaml"],
    data_files=[('', ['requirements.txt', 'pytest.ini', '.coveragerc', 'LICENSE', 'README.md', 'CHANGELOG.md'])]
)



from typing import Dict, List

import pytest

from json_to_models.dynamic_typing import (DDict, DList, DOptional, DUnion, FloatString, IntString, ModelMeta,
                                           compile_imports)
from json_to_models.models.attr import AttrsModelCodeGenerator, DEFAULT_ORDER
from json_to_models.models.base import METADATA_FIELD_NAME, generate_code, sort_kwargs
from json_to_models.models.structure import sort_fields
from test.test_code_generation.test_models_code_generator import model_factory, trim


def test_attrib_kwargs_sort():
    sorted_kwargs = sort_kwargs(dict(
        y=2,
        metadata='b',
        converter='a',
        default=None,
        x=1,
    ), DEFAULT_ORDER)
    expected = ['default', 'converter', 'y', 'x', 'metadata']
    for k1, k2 in zip(sorted_kwargs.keys(), expected):
        assert k1 == k2
    try:
        sort_kwargs({}, ['wrong_char'])
    except ValueError as e:
        assert e.args[0].endswith('wrong_char')
    else:
        assert 0, "XPass"


def field_meta(original_name):
    return f"metadata={{'{METADATA_FIELD_NAME}': '{original_name}'}}"


# Data structure:
# pytest.param id -> {
#   "model" -> (model_name, model_metadata),
#   test_name -> expected, ...
# }
test_data = {
    "base": {
        "model": ("Test", {
            "foo": int,
            "Bar": int,
            "baz": float
        }),
        "fields_data": {
            "foo": {
                "name": "foo",
                "type": "int",
                "body": f"attr.ib()"
            },
            "Bar": {
                "name": "bar",
                "type": "int",
                "body": f"attr.ib({field_meta('Bar')})"
            },
            "baz": {
                "name": "baz",
                "type": "float",
                "body": f"attr.ib()"
            }
        },
        "fields": {
            "imports": "",
            "fields": [
                f"foo: int = attr.ib()",
                f"bar: int = attr.ib({field_meta('Bar')})",
                f"baz: float = attr.ib()",
            ]
        },
        "generated": trim(f"""
        import attr
        
        
        @attr.s
        class Test:
            foo: int = attr.ib()
            bar: int = attr.ib({field_meta('Bar')})
            baz: float = attr.ib()
        """)
    },
    "complex": {
        "model": ("Test", {
            "foo": int,
            "baz": DOptional(DList(DList(str))),
            "bar": DOptional(IntString),
            "qwerty": FloatString,
            "asdfg": DOptional(int),
            "dict": DDict(int),
            "ID": int,
            "not": bool,
            "1day": int,
            "день_недели": str,
        }),
        "fields_data": {
            "foo": {
                "name": "foo",
                "type": "int",
                "body": "attr.ib()"
            },
            "baz": {
                "name": "baz",
                "type": "Optional[List[List[str]]]",
                "body": "attr.ib(factory=list)"
            },
            "bar": {
                "name": "bar",
                "type": "Optional[IntString]",
                "body": "attr.ib(default=None, converter=optional(IntString))"
            },
            "qwerty": {
                "name": "qwerty",
                "type": "FloatString",
                "body": "attr.ib(converter=FloatString)"
            },
            "asdfg": {
                "name": "asdfg",
                "type": "Optional[int]",
                "body": "attr.ib(default=None)"
            },
            "dict": {
                "name": "dict_",
                "type": "Dict[str, int]",
                "body": f"attr.ib({field_meta('dict')})"
            },
            "ID": {
                "name": "id_",
                "type": "int",
                "body": f"attr.ib({field_meta('ID')})"
            },
            "not": {
                "name": "not_",
                "type": "bool",
                "body": f"attr.ib({field_meta('not')})"
            },
            "1day": {
                "name": "one_day",
                "type": "int",
                "body": f"attr.ib({field_meta('1day')})"
            },
            "день_недели": {
                "name": "den_nedeli",
                "type": "str",
                "body": f"attr.ib({field_meta('день_недели')})"
            }
        },
        "generated": trim(f"""
        import attr
        from json_to_models.dynamic_typing import FloatString, IntString
        from json_to_models.models import ClassType
        from json_to_models.models.string_converters import convert_strings
        from typing import Dict, List, Optional


        @attr.s
        @convert_strings(['bar#O.S', 'qwerty'], class_type=ClassType.Attrs)
        class Test:
            foo: int = attr.ib()
            qwerty: FloatString = attr.ib()
            dict_: Dict[str, int] = attr.ib({field_meta('dict')})
            id_: int = attr.ib({field_meta('ID')})
            not_: bool = attr.ib({field_meta('not')})
            one_day: int = attr.ib({field_meta('1day')})
            den_nedeli: str = attr.ib({field_meta('день_недели')})
            baz: Optional[List[List[str]]] = attr.ib(factory=list)
            bar: Optional[IntString] = attr.ib(default=None)
            asdfg: Optional[int] = attr.ib(default=None)
        """)
    },
    "converters": {
        "model": ("Test", {
            "a": int,
            "b": IntString,
            "c": DOptional(FloatString),
            "d": DList(DList(DList(IntString))),
            "e": DDict(IntString),
            "u": DUnion(DDict(IntString), DList(DList(IntString))),
        }),
        "generated": trim("""
        import attr
        from json_to_models.dynamic_typing import FloatString, IntString
        from json_to_models.models import ClassType
        from json_to_models.models.string_converters import convert_strings
        from typing import Dict, List, Optional, Union


        @attr.s
        @convert_strings(['b', 'c#O.S', 'd#L.L.L.S', 'e#D.S'], class_type=ClassType.Attrs)
        class Test:
            a: int = attr.ib()
            b: IntString = attr.ib()
            d: List[List[List[IntString]]] = attr.ib()
            e: Dict[str, IntString] = attr.ib()
            u: Union[Dict[str, IntString], List[List[IntString]]] = attr.ib()
            c: Optional[FloatString] = attr.ib(default=None)
        """)
    }
}

test_data_unzip = {
    test: [
        pytest.param(
            model_factory(*data["model"]),
            data[test],
            id=id
        )
        for id, data in test_data.items()
        if test in data
    ]
    for test in ("fields_data", "fields", "generated")
}


@pytest.mark.parametrize("value,expected", test_data_unzip["fields_data"])
def test_fields_data_attr(value: ModelMeta, expected: Dict[str, dict]):
    gen = AttrsModelCodeGenerator(value, meta=True)
    required, optional = sort_fields(value)
    for is_optional, fields in enumerate((required, optional)):
        for field in fields:
            field_imports, data = gen.field_data(field, value.type[field], bool(is_optional))
            assert data == expected[field]


@pytest.mark.parametrize("value,expected", test_data_unzip["fields"])
def test_fields_attr(value: ModelMeta, expected: dict):
    expected_imports: str = expected["imports"]
    expected_fields: List[str] = expected["fields"]
    gen = AttrsModelCodeGenerator(value, meta=True)
    imports, fields = gen.fields
    imports = compile_imports(imports)
    assert imports == expected_imports
    assert fields == expected_fields


@pytest.mark.parametrize("value,expected", test_data_unzip["generated"])
def test_generated_attr(value: ModelMeta, expected: str):
    generated = generate_code(([{"model": value, "nested": []}], {}), AttrsModelCodeGenerator,
                              class_generator_kwargs={'meta': True, 'post_init_converters': True})
    assert generated.rstrip() == expected, generated



import pytest

from json_to_models.dynamic_typing import (BooleanString, DDict, DList, DOptional, DTuple, DUnion, FloatString,
                                           IntString, Null, Unknown)
from json_to_models.generator import MetadataGenerator

# MetaData | Optimized MetaData
test_data = [
    pytest.param(DUnion(int), int, id="single_DUnion"),
    pytest.param(
        {'a': DUnion({'b': int}, {'b': float})},
        {'a': {'b': float}},
        id="merge_nested_dicts"
    ),
    pytest.param(
        {'1': DUnion(
            {'a': DUnion({'b': int}, {'b': float})},
            {'a': DUnion({'b': float}, {'b': int})},
            {'a': Null},
        )},
        {'1': {'a': DOptional({'b': float})}},
        id="merge_nested_dicts"
    ),
    pytest.param(
        DUnion(FloatString, IntString),
        FloatString,
        id="str_types_merge"
    ),
    pytest.param(
        DUnion(FloatString, BooleanString),
        str,
        id="str_types_merge_not_resolved"
    ),
    pytest.param(
        DList(DUnion(FloatString, BooleanString)),
        DList(str),
        id="str_types_merge_not_resolved_simple_wrapper"
    ),
    pytest.param(
        DTuple(DUnion(FloatString, BooleanString), int),
        DTuple(str, int),
        id="str_types_merge_not_resolved_complex_wrapper"
    ),
    pytest.param(
        DUnion(FloatString, str),
        str,
        id="union_str_str_based"
    ),
    pytest.param(
        DUnion(DList(int), DList(str), str),
        DUnion(DList(DUnion(int, str)), str),
        id="union_of_lists"
    ),
    pytest.param(
        DUnion(Unknown, str),
        str,
        id="unknown_vs_single"
    ),
    pytest.param(
        DUnion(Unknown, str, int),
        DUnion(str, int),
        id="unknown_vs_multi"
    ),
    pytest.param(
        DUnion(DList(Unknown), DList(str)),
        DList(str),
        id="list_unknown_vs_single"
    ),
    pytest.param(
        DUnion(DList(Unknown), DList(str), DList(int)),
        DList(DUnion(str, int)),
        id="list_unknown_vs_multi"
    ),
    pytest.param(
        DUnion(Null, str),
        DOptional(str),
        id="optional_str"
    ),
    pytest.param(
        DUnion(Null, DList(str)),
        DOptional(DList(str)),
        id="optional_list"
    ),
    pytest.param(
        DList(DUnion(Null, str)),
        DList(DOptional(str)),
        id="list_of_optional_strings"
    ),
    pytest.param(
        DUnion(Null, DList(str), int),
        DOptional(DUnion(DList(str), int)),
        id="optional_list_or_int"
    ),
    pytest.param(
        DUnion(Null, DList(DUnion(str, int))),
        DOptional(DList(DUnion(str, int))),
        id="optional_list_of_str_or_int"
    ),
    pytest.param(
        DList(DUnion(Null, str, int)),
        DList(DOptional(DUnion(str, int))),
        id="list_of_optional_strings_ot_int"
    ),
    pytest.param(
        DList(DUnion(str, int, FloatString, IntString)),
        DList(DUnion(str, int)),
        id="union_of_str_int_FloatString"
    ),
    pytest.param(
        DOptional(DUnion(DOptional(str), str)),
        DOptional(str),
        id="optional_union_nested"
    ),
    pytest.param(
        DUnion(Null, str, Null),
        DOptional(str),
        id="optional_str"
    ),
    pytest.param(
        DUnion(DDict(str), DDict(str), DDict(str)),
        DDict(str),
        id="dict_union"
    ),
    pytest.param(
        DUnion(DDict(str), DDict(int), DDict(str)),
        DDict(DUnion(str, int)),
        id="dict_union_2"
    ),
]


@pytest.mark.parametrize("value,expected", test_data)
def test_optimize_type(models_generator: MetadataGenerator, value, expected):
    result = models_generator.optimize_type(value)
    assert result == expected
